/*
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2012 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "timer.h"
#include "rodent.h"
#include "data.h"
#include "parser.h"
#include "selector.h"
#include "trans.h"
#include "search/search.h"
#include "move/move.h"
#include "book.h"

#define DELETE_MOVE 888

void sBook::Init(sPosition * p) {

	 Timer.SetStartTime();
	 nOfRecords = 0;
	 nOfGuideRecords = 0;
	 ReadOwnBookFile("bigbook.wtf"); // read main book
	 ReadTextFileToGuideBook(p, "guidebook.txt", NO_CL);
}

void sBook::BookDoctor(sPosition * p) {
	 int move;
	 UNDO u;

	 FILE *doctor_file; 
     doctor_file = fopen("doctor.txt","a+"); 

	 int flagIsProblem = 0;
	 printf("Book doctor launched. It looks for a line that might need Your decision and saves it to doctor.txt\n");
	 
	 for (int i = 0; i < 10000; i++ ) 
	 {
		 SetPosition(p, START_POS);
		 printf("\n");
		 fprintf(doctor_file, "\n");
        for (;;)
		{
		 move = GetBookMove(p, 0, &flagIsProblem);
		 if (move) { 
			 Manipulator.DoMove(p, move, &u);
             if (flagIsProblem) break; 
	         printf("%c",File(Fsq(move)) + 'a');
			 fprintf(doctor_file, "%c", File(Fsq(move)) + 'a');
	         printf("%c",Rank(Fsq(move)) + '1');
			 fprintf(doctor_file, "%c", Rank(Fsq(move)) + '1');
             printf("%c",File(Tsq(move)) + 'a');
			 fprintf(doctor_file, "%c", File(Tsq(move)) + 'a');
	         printf("%c",Rank(Tsq(move)) + '1');
			 fprintf(doctor_file, "%c", Rank(Tsq(move)) + '1');
	         printf(" ");
			 fprintf(doctor_file, " ");
		 }
		 else break;
		}
		if (flagIsProblem) break;
	 }

     fclose(doctor_file);

}

int sBook::GetBookMove(sPosition *p, int canPrint, int *flagIsProblem) {
	int i;
	int floorVal   = 0;
	int curVal     = 0;
	int bestVal    = 0;
	int choice     = 0;
	int maxFreq    = 0;
	int values[100];
	int searchScores[100];
	U64 localHash = GetBookHash(p);

	nOfChoices = 0;

	// no book move returned in analyse mode
	if (Data.isAnalyzing) return 0;

	for (i = 0; i < nOfGuideRecords; i++ ) {
		if (guideBook[i].hash == localHash
		&& IsLegal(p, guideBook[i].move ) )
		{
			moves[nOfChoices]  = guideBook[i].move;
			if (guideBook[i].freq > 0 ) values[nOfChoices] = guideBook[i].freq + 20;
			else                        values[nOfChoices] = -1;
	       // we add 20, so that any move not marked as bad in the text-based
		   // opening book has a chance to be played
			nOfChoices++;
		}
	}

	if (nOfChoices) {

       srand(Timer.GetMS() );

	   if (canPrint) printf("info string ");
	   for (i = 0; i < nOfChoices; i++ ) {
		   
		   // display info about possible choices
		   if (canPrint) printf(" ");
		   MoveToStr(moves[i], testString);
		   printf( testString );
		   if (values[i] > 0 ) { 
			   if (canPrint) printf(" %d ", values[i] );
		   }
		   else              if (canPrint)  printf("? ");

           // pick move with the best random value based on frequency
		   if (values[i] > 0) curVal = 1 + rand() % values[i]; 
		   else curVal = -1;

		   if (curVal > bestVal) 
		   {
			   bestVal = curVal;
			   choice  = i;
		   }
	   }
	   if (canPrint) printf(" guide \n");

       return moves[choice];
	}

	// find a convenient starting point to avoid looping through entire book
	int iStart  = 0;
	i = 1;
	if (nOfRecords > 10000) 
	{
	   for(;;) {
         if (myBook[i*10000].hash < localHash) iStart = i*10000;
	     else break;         
	     i++;
	     if (i*10000 > nOfRecords) break;
	   }
	}

	// find possible book moves
	for (i = iStart; i < nOfRecords; i++ ) 
	{
        if (myBook[i].hash > localHash) break;

		if (myBook[i].hash == localHash
		&& IsLegal(p, myBook[i].move ) ){
			moves[nOfChoices]  = myBook[i].move;
			values[nOfChoices] = myBook[i].freq;
			nOfChoices++;
		}
	}

	if (nOfChoices) {

       // get maximum frequency of a move - it will be used
       // to filter out moves that were played not often enough
	   for (i = 0; i < nOfChoices; i++ )
	   {
          if (values[i] > maxFreq ) maxFreq = values[i];
	   }

       srand(Timer.GetMS() );
	   if (Data.verifyBook) floorVal = Searcher.VerifyValue(p, 9, 0);

	   for (i = 0; i < nOfChoices; i++ ) {
		   
		   // display info about possible choices
		   if (canPrint) printf("info string ");
		   MoveToStr(moves[i], testString);
		   if (canPrint || (values[i] > 0 && IsInfrequent(values[i], maxFreq) ) ) printf( testString );

		   if (values[i] > 0 ) 
		   { 
			   // we are in the special mode, in which Rodent takes some time
			   // before picking a book move
			   if (Data.verifyBook)
			   {
   			     // search to get move value if a move won't be excluded due to low frequency
                 // and if its frequency isn't exceptionally high
				 if ( !IsInfrequent(values[i], maxFreq) )
				 {
					  if (values[i] > 5000) searchScores[i] = maxFreq;
				      else                  searchScores[i] = Searcher.VerifyValue(p, 8, moves[i]);
				 }
				 else 
					  searchScores[i] = 0;

				 // print debug book information
			     if (canPrint) printf( " freq %4d floor %4d val %4d delta %4d ", values[i], floorVal, searchScores[i], searchScores[i] - floorVal );
			     
				 // we're in a bad position in first place (just debug info)
				 if ( searchScores[i] < -50 && canPrint ) printf("bad ");

				 // exclude rare moves if there is a popular alternative
				 if ( IsInfrequent(values[i], maxFreq) ) 
				 {
					 printf(" infrequent ");
					 *flagIsProblem = 1;
					 values[i] = -100;
				 }

			     // this move causes score drop, so we avoid it
				 if ( searchScores[i] - floorVal < -50 ) 
				 { 
					 if (canPrint) printf("drop ");
					 values[i] = -100;
				 }

			     // get search-based move value
				 if (values[i] > 0) values[i] = 50 + searchScores[i] - floorVal;
			   }

			   // book move choice based on frequency, without verification search
			   else 
			   {
				   if ( IsInfrequent(values[i], maxFreq) )
				   {
					   values[i] = 0;
					   printf(" infrequent ");
					   *flagIsProblem = 1;
				   }
				   if (canPrint) printf(" %d ", values[i] );
			   }
		   }
		   else                if (canPrint) printf("? ");
		   
           // pick move randomly, based on its value
		   // (we add 5 to ensure choice between equally rare moves)
		   if (values[i] > 0) curVal = 1 + rand() % (values[i] + 5); 
		   else curVal = -1;
		   if (curVal > bestVal) 
		   {
			   bestVal = curVal;
			   choice  = i;
		   }
		   if (canPrint) printf("\n");
	   }

      if ( PrintMissingMoves(p) ) *flagIsProblem = 1;
      if (bestVal > 0) return moves[choice];
	}

	if ( PrintMissingMoves(p) ) *flagIsProblem = 1;
    return 0;
}


int sBook::PrintMissingMoves(sPosition *p)
{
      // detection of non-book moves leading to book positions
      // TODO: separate function
      sSelector Selector;
	  UNDO undoData[1];
	  int move = 0;
	  int flagMoveType;
	  int isProblem = 0;

      Selector.InitMoves(p, move, MAX_PLY);

      // LOOP THROUGH THE MOVE LIST
      while ( move = Selector.NextMove(0, &flagMoveType) ) {

	     // MAKE A MOVE
	     Manipulator.DoMove(p, move, undoData);    
	
	     // UNDO ILLEGAL MOVES
	     if (IllegalPosition(p)) 
	     { 
		    Manipulator.UndoMove(p, move, undoData); 
		    continue; 
	     }
     
		 for (int i = 0; i < nOfRecords; i++) 
		 {
		   if (myBook[i].hash == GetBookHash(p) ) {

           int isUsed = 0;
		   for (int j = 0; j < nOfChoices; j++) 
		   {
			   if (moves[j] == move) isUsed = 1;
		   }

		   if (!isUsed) 
		   {
		      printf("info string missing ");
			  isProblem = 1;
		      MoveToStr(move, testString);
		      printf( testString );
		      printf("\n");
		   }
		   break;
			 }
		 }

	     Manipulator.UndoMove(p, move, undoData); 
  }

  return isProblem;
}

void sBook::AddMoveToGuideBook(U64 hashKey, int move, int val) {

    // if move is already in the book, just change its frequency 
	for (int i = 0; i < nOfGuideRecords; i++ ) {
         if ( guideBook[i].hash == hashKey 
		 &&   guideBook[i].move == move) 
		 {
			  guideBook[i].freq += val;	 
			  return;
		 }
	 }  

	 // otherwise save it in the last slot
	 guideBook[nOfGuideRecords].hash = hashKey;
	 guideBook[nOfGuideRecords].move = move;
	 guideBook[nOfGuideRecords].freq = val;
	 nOfGuideRecords++;
}

void sBook::AddMoveToMainBook(U64 hashKey, int move, int val) {

    // if move is already in the book, just change its frequency 
	for (int i = 0; i < nOfRecords; i++ ) 
	{
         if ( myBook[i].hash == hashKey 
		 &&   myBook[i].move == move) 
		 {
			 if (val == DELETE_MOVE) myBook[i].hash = 0;
			 if (myBook[i].freq <= 20000 || val != 1) myBook[i].freq += val;
			 return;
		 }
	 }  

	 // otherwise save it in the last slot
	 if (val != DELETE_MOVE)
	 {
	    myBook[nOfRecords].hash = hashKey;
	    myBook[nOfRecords].move = move;
	    myBook[nOfRecords].freq = val;
	 }
	 nOfRecords++;
}

int sBook::IsMoveInBook(U64 hashKey, int move)
{
	printf("nofrecords %d", nOfRecords);
	if (nOfRecords == 0) return 0;

	for (int i = 0; i < nOfRecords; i++ ) 
	{
		 if ( i % 1000 == 0) printf (" %d ", i); // fixing a heisenbug
         if ( myBook[i].hash == hashKey 
			 &&   myBook[i].move == move) {
		 return 1;
		 }
	}
	return 0;
}

void sBook::AddLineToGuideBook(sPosition *p, char *ptr, int excludedColor)
{
  char token[512];
  UNDO u[1];
  int move;
  int freq;
    
  SetPosition(p, START_POS);

  for (;;) 
  {
    ptr = Parser.ParseToken(ptr, token);
	  
    if (*token == '\0') break;

	move = StrToMove(p, token);
      
	if (IsLegal(p, move)) 
	{
		// apply move frequency modifiers
		freq = 1;
        if (strstr(token, "?")) freq = -100;
		if (strstr(token, "!")) freq = +100;
        if (strstr(token, "??")) freq = -4900;
		if (strstr(token, "!!")) freq = +900;

		// add move to book if we accept moves of a given color
		if (p->side != excludedColor)
		   AddMoveToGuideBook( GetBookHash(p), move, freq);
        
		Manipulator.DoMove(p, move, u);
	}
	  else break;

    if (p->reversibleMoves == 0)
        p->head = 0;
    }
}

void sBook::AddLineToMainBook(sPosition *p, char *ptr, int excludedColor)
{
  char token[2048];
  UNDO u[1];
  int move;
  int freq;
    
  SetPosition(p, START_POS);

  for (;;) 
  {
    ptr = Parser.ParseToken(ptr, token);
	  
    if (*token == '\0') break;

	move = StrToMove(p, token);
      
	if (IsLegal(p, move)) 
	{
		// apply move frequency modifiers
		freq = 1;
        if (strstr(token, "?")) freq  = -100;
		if (strstr(token, "!")) freq  = +100;
        if (strstr(token, "??")) freq = -4900;
		if (strstr(token, "!!")) freq = +900;
		if (strstr(token, "xx")) freq = DELETE_MOVE;

		// disabled code for book feeder
		/**
        if (freq == 1 && !IsMoveInBook( GetBookHash(p), move ) && p->side != excludedColor )
		{
			if ( Searcher.VerifyValue(p, 16, move) < -75 ) freq = -100;
		}
		/**/

		// add move to book if we accept moves of a given color
		if (p->side != excludedColor)
		   AddMoveToMainBook( GetBookHash(p), move, freq);
        
		Manipulator.DoMove(p, move, u);
	}
	  else break;

    if (p->reversibleMoves == 0)
        p->head = 0;
    }
}


void sBook::ReadTextFileToGuideBook(sPosition *p, char *fileName, int excludedColor)
{
	  FILE *book_file; 
	  char line[256];
	  int line_no = 0;

      // exit if book file doesn't exist
	  if ( (book_file = fopen(fileName, "r")) == NULL ) 
		 return;

      nOfGuideRecords = 0;

      // process book file line by line 
	  while ( fgets(line, 250, book_file) ) {
		    ++line_no;
		    if(line[0] == ';') continue; // don't process comment lines
		    AddLineToGuideBook(p, line, excludedColor);
	  }
	  fclose(book_file);
 }

void sBook::SplitContinousBookFormat(char *fileName)
{
	  FILE *book_file; 
	  FILE *out_file;
	  char line[256];
	  int line_no = 0;

      // exit if book file doesn't exist
	  if ( (book_file = fopen(fileName, "r")) == NULL ) 
		 return;

	  out_file = fopen("loose.txt","w"); 

      // process book file line by line 
	  while ( fgets(line, 250, book_file) ) {
		    ++line_no;
		    if(line[0] == ';') continue; // don't process comment lines
		    int length=strlen(line);
			for (int i = 0; i < length;  i++) 
			{
				printf("%c", line[i]);
				fprintf(out_file, "%c", line[i]);
				if ( (i+1) % 4 == 0) fprintf(out_file," ");
			}
	  }

	  fclose(book_file);
	  fclose(out_file);
 }

void sBook::ReadMainBookFromOwnFile(sPosition *p, char *fileName, int excludedColor)
{
	  FILE *book_file; 
	  char line[2048];
	  int line_no = 0;

      // exit if book file doesn't exist
	  if ( (book_file = fopen(fileName, "r")) == NULL ) 
		 return;

      // process book file line by line 
	  while ( fgets(line, 2040, book_file) ) {
		    ++line_no;
			if ( line_no % 100 == 0 ) printf("Adding line no. %d\r",line_no);
		    if(line[0] == ';') continue; // don't process comment lines
		    AddLineToMainBook(p, line, excludedColor);
	  }

	  printf("Adding line no. %d\r",line_no);
	  fclose(book_file);
 }


U64 sBook::GetBookHash(sPosition *p) {
	// TODO: return Polyglot-compatibile hash key
	U64 bookKey = p->hashKey / 4;
	if (bookKey < 0) bookKey *= -1;
    return (signed long long) (bookKey);
}

void sBook::SaveBookInOwnFormat(char *fileName) {
	 FILE *book_file; 

     book_file = fopen(fileName,"a+"); 
	 for (int i = 0; i < nOfRecords; i++ ) 
	 {
		 if (myBook[i].hash != 0)
		 fprintf(book_file, "%I64u, %d, %d \n", myBook[i].hash, myBook[i].move, myBook[i].freq);
	 }
     fclose(book_file);
}

void sBook::ReadOwnBookFile(char *fileName)
{
	  FILE *book_file; 
	  char line[256];
	  nOfRecords = 0;

      // exit if book file doesn't exist...
	  if ( (book_file = fopen(fileName, "r")) == NULL ) 
		 return;

      // ...process book file line by line 
	  while ( fgets(line, 250, book_file) ) 
	  {
		  ParseBookEntry(line, nOfRecords);
		  ++nOfRecords;
	  }

	  fclose(book_file);
 }

void sBook::ParseBookEntry(char * ptr, int line_no) {

  char token[256];
  int token_no = 1;

  for (;;) 
  {
    ptr = Parser.ParseToken(ptr, token);
	if (token_no == 1) myBook[line_no].hash = atoull(token);
	if (token_no == 2) myBook[line_no].move = atoi(token); 
	if (token_no == 3) myBook[line_no].freq = atoi(token); 
	token_no++;
	  
    if (*token == '\0') break;
  }

}

void sBook::SortMainBook(void) {
	int i, j, change;
	sBookEntry tmp;
 
  for (i=0; i<nOfRecords-1; ++i) 
  {
       if (i % 100 == 0 ) printf ("%d   \r", i);

       change=0;
       for (j=0; j<nOfRecords-1-i; j++)
       { 
		   if (myBook[j+1].hash < myBook[j].hash // comparing neighbours
		   || (myBook[j+1].hash == myBook[j].hash && myBook[j+1].freq < myBook[j].freq ) 
		   )   
          {  
              tmp = myBook[j];      
              myBook[j] = myBook[j+1];
              myBook[j+1] = tmp;    // bubble goes up     
              change=1;
          }
       }
       if(!change) break;      // no changes - book is sorted

       for (j=nOfRecords-2; j>1; j--)
       { 
		   if (myBook[j+1].hash < myBook[j].hash // comparing neighbours
		   || (myBook[j+1].hash == myBook[j].hash && myBook[j+1].freq < myBook[j].freq ) )   
          {  
              tmp = myBook[j];      
              myBook[j] = myBook[j+1];
              myBook[j+1] = tmp;    // bubble goes down     
              change=1;
          }
       }
  }

};

void sBook::FeedMainBook() 
{
	 sPosition * p;
	 printf("Feeding book moves for both sides\n");
	 ReadMainBookFromOwnFile(p, "feed.txt", NO_CL);
	 printf("\nFeeding book moves for white only\n");
	 ReadMainBookFromOwnFile(p, "feed_white.txt", BLACK);
	 printf("\nFeeding book moves for black only\n");
	 ReadMainBookFromOwnFile(p, "feed_black.txt", WHITE);
	 printf("\nSorting book moves\n");
	 SortMainBook();
	 SaveBookInOwnFormat("newbook.wtf");
}

// Arbitrary definition of an infrequent move: less than 100 entries
// in the book and less than 15% of frequency of the most popular move
// in a given position.
int sBook::IsInfrequent(int val, int maxFreq)
{
return ( val < 100 && val < maxFreq / 15); 
}